#ifndef TESTLIB10_INTERFACE_CUSTOM_C
#  error "TESTLIB10_INTERFACE_CUSTOM_C incorrectly not defined!"
#endif

#ifndef TESTLIB10_INTERFACE_CUSTOM_L
#  error "TESTLIB10_INTERFACE_CUSTOM_L incorrectly not defined!"
#endif

int testLib10(void);

int main(void)
{
  return testLib10();
}
