#ifdef STATICLIB2_IFACE_1
#  error "STATICLIB2_IFACE_1 incorrectly defined"
#endif
#ifdef STATICLIB2_IFACE_2
#  error "STATICLIB2_IFACE_2 incorrectly defined"
#endif

int staticlib1()
{
  return 0;
}
