extern void order_I_poison(void);

void order_A(void)
{
  order_I_poison();
}

void order_B(void)
{
  order_I_poison();
}

void order_C(void)
{
  order_I_poison();
}

void order_D(void)
{
  order_I_poison();
}

void order_E(void)
{
  order_I_poison();
}

void order_F(void)
{
  order_I_poison();
}

void order_G(void)
{
  order_I_poison();
}

void order_H(void)
{
  order_I_poison();
}
