#ifndef OwnDotUnderscore_P_HPP
#define OwnDotUnderscore_P_HPP

#include <QObject>

class OwnDotUnderscorePrivate : public QObject
{
  Q_OBJECT
public:
  OwnDotUnderscorePrivate();
  ~OwnDotUnderscorePrivate();
};

#endif
