#ifndef STYLEB_HPP
#define STYLEB_HPP

#include <QObject>

/* clang-format off */
/// Q_OBJECT behind a brace on a new line
///
class StyleB : public QObject
{ Q_OBJECT
public:
  StyleB();
};
/* clang-format on */

#endif
