#include <genex_config.h>
/* FIXME Get this working */
/*#include <intdir_config.h>*/

const char* generatorobj_genex_config_definition(void)
{
  return GENEX_CONFIG_DEFINITION;
}

const char* generatorobj_genex_config_include_dir(void)
{
  return GENEX_CONFIG_INCLUDE_DIR;
}

/* FIXME Get this working */
/*const char* generatorobj_intdir_config_definition(void)
{
  return INTDIR_CONFIG_DEFINITION;
}

const char* generatorobj_intdir_config_include_dir(void)
{
  return INTDIR_CONFIG_INCLUDE_DIR;
}*/
