#ifndef CUSTOM_A_IFACE1
#  error "CUSTOM_A_IFACE1 incorrectly not defined"
#endif

#ifndef CUSTOM_A_IFACE2
#  error "CUSTOM_A_IFACE2 incorrectly not defined"
#endif

#ifndef CUSTOM_A_OBJECT1
#  error "CUSTOM_A_OBJECT1 incorrectly not defined"
#endif

#ifndef CUSTOM_A_TARGET_NAME_OBJECT1
#  error "CUSTOM_A_TARGET_NAME_OBJECT1 incorrectly not defined"
#endif

#ifndef CUSTOM_A_TARGET_TYPE_OBJECT_LIBRARY
#  error "CUSTOM_A_TARGET_TYPE_OBJECT_LIBRARY incorrectly not defined"
#endif

#ifndef CUSTOM_C_IFACE1
#  error "CUSTOM_C_IFACE1 incorrectly not defined"
#endif

#ifndef CUSTOM_C_OBJECT1
#  error "CUSTOM_C_OBJECT1 incorrectly not defined"
#endif

#ifdef CUSTOM_C_OBJECT1_IFACE
#  error "CUSTOM_C_OBJECT1_IFACE incorrectly defined"
#endif

#ifndef CUSTOM_U_IFACE1
#  error "CUSTOM_U_IFACE1 incorrectly not defined"
#endif

#ifndef CUSTOM_U_IFACE2
#  error "CUSTOM_U_IFACE2 incorrectly not defined"
#endif

#ifndef CUSTOM_U_OBJECT1
#  error "CUSTOM_U_OBJECT1 incorrectly not defined"
#endif

#ifndef CUSTOM_U_TARGET_NAME_OBJECT1
#  error "CUSTOM_U_TARGET_NAME_OBJECT1 incorrectly not defined"
#endif

#ifndef CUSTOM_U_TARGET_TYPE_OBJECT_LIBRARY
#  error "CUSTOM_U_TARGET_TYPE_OBJECT_LIBRARY incorrectly not defined"
#endif

#ifndef CUSTOM_W_IFACE1
#  error "CUSTOM_W_IFACE1 incorrectly not defined"
#endif

#ifndef CUSTOM_W_OBJECT1
#  error "CUSTOM_W_OBJECT1 incorrectly not defined"
#endif

int object1(void)
{
  return 0;
}
