/* Distributed under the OSI-approved BSD 3-Clause License.  See accompanying
   file Copyright.txt or https://cmake.org/licensing for details.  */
#ifndef BLUB_H
#define BLUB_H

#include <QObject>

class Blub
{
public:
  Blub();
  void blubber();
};

#endif
