#ifndef SCOPED_A_1
#  error "SCOPED_A_1 not defined"
#endif
#ifndef SCOPED_B_1
#  error "SCOPED_B_1 not defined"
#endif
#ifndef SCOPED_A_2
#  error "SCOPED_A_2 not defined"
#endif
void Scope(void)
{
}
