#ifndef OtherUnderscore_P_HPP
#define OtherUnderscore_P_HPP

#include <QObject>

class OtherUnderscorePrivate : public QObject
{
  Q_OBJECT
public:
  OtherUnderscorePrivate();
  ~OtherUnderscorePrivate();
};

#endif
