class HelloWorld2
{
    public static void main(String args[])
    {
        B b;
        b = new B();
        b.printName();

        C c;
        c = new C();
        c.printName();

        System.out.println("Hello World!");
    }
}
