
#ifndef HAVE_PACK1
#  error Expected HAVE_PACK1
#endif

#ifndef HAVE_PACK2
#  error Expected HAVE_PACK2
#endif

#ifndef HAVE_PACK3
#  error Expected HAVE_PACK3
#endif

#ifndef HAVE_PACK4
#  error Expected HAVE_PACK4
#endif

#ifndef HAVE_PACK5
#  error Expected HAVE_PACK5
#endif

#ifndef HAVE_PACK6
#  error Expected HAVE_PACK6
#endif

#ifndef HAVE_PACK7
#  error Expected HAVE_PACK7
#endif

#ifndef HAVE_PACK7_COMP1
#  error Expected HAVE_PACK7_COMP1
#endif

#ifndef HAVE_PACK8
#  error Expected HAVE_PACK8
#endif

int main(int argc, char** argv)
{
  return 0;
}
