/* Distributed under the OSI-approved BSD 3-Clause License.  See accompanying
   file Copyright.txt or https://cmake.org/licensing for details.  */
#pragma once

enum class codecvt_Encoding
{
  None,
  UTF8,
  UTF8_WITH_BOM,
  ANSI,
  ConsoleOutput,
};
